(include-book "models/jvm/m1/m1" :dir :system)
(in-package "M1")

(defun execute-ILOAD (inst s)   ; (ILOAD n)
      (make-state (+ 1 (pc s))
                  (locals s)
                  (push (nth (arg1 inst) (locals s))
                        (stack s))
                  (program s)))
(defun execute-ISTORE (inst s)  ; (ISTORE n)
      (make-state (+ 1 (pc s))
                  (update-nth (arg1 inst)
                              (top (stack s))
                              (locals s))
                  (pop (stack s))
                  (program s)))

(defun execute-IADD (inst s)  ; (IADD)
      (declare (ignore inst))
      (make-state (+ 1 (pc s))
                  (locals s)
                  (push (+ (top (pop (stack s)))
                           (top (stack s)))
                        (pop (pop (stack s))))
                  (program s)))

(defun execute-IFEQ (inst s)  ; (IFEQ offset)
      (make-state (if (equal (top (stack s)) 0)
                      (+ (arg1 inst) (pc s))
                      (+ 1 (pc s)))
                  (locals s)
                  (pop (stack s))
                  (program s)))

(defun do-inst (inst s)
  (if (equal (op-code inst) 'ILOAD)  (execute-ILOAD inst s)
  (if (equal (op-code inst) 'ICONST) (execute-ICONST inst s)
  (if (equal (op-code inst) 'IADD)   (execute-IADD inst s)
  (if (equal (op-code inst) 'ISUB)   (execute-ISUB inst s)
  (if (equal (op-code inst) 'IMUL)   (execute-IMUL inst s)
  (if (equal (op-code inst) 'ISTORE) (execute-ISTORE inst s)
  (if (equal (op-code inst) 'GOTO)   (execute-GOTO inst s)
  (if (equal (op-code inst) 'IFEQ)   (execute-IFEQ inst s)
                                  s)))))))))
(defun next-inst (s)
      (nth (pc s) (program s)))

(defun step (s)
      (do-inst (next-inst s) s))
(defun m1 (s n)
      (if (zp n)
          s
          (m1 (step s) (- n 1))))

(quote (back to talk))

(defconst *pi*
       '((ICONST 0)   ; 0
         (ISTORE 2)   ; 1  a $=$ 0;
         (ILOAD 0)    ; 2  [loop:]
         (IFEQ 10)    ; 3  if x$=$0 then go to end;
         (ILOAD 0)    ; 4
         (ICONST 1)   ; 5
         (ISUB)       ; 6
         (ISTORE 0)   ; 7  x $=$ x-1;
         (ILOAD 1)    ; 8
         (ILOAD 2)    ; 9
         (IADD)       ;10
         (ISTORE 2)   ;11  a $=$ y+a;
         (GOTO -10)   ;12  go to loop
         (ILOAD 2)    ;13  [end:]
         (HALT)       ;14 ``return'' a
         ))

(defconst *si*
      (make-state 0               ; initial pc
                  '(5 7 17 19 23) ; initial locals
                  nil             ; initial stack
                  *pi*            ; program
                  ))

(m1 *si* 100)

(quote (back to talk))

(defun loop-clk (x)
         (if (zp x)
             3
             (clk+ 11
                   (loop-clk (- x 1)))))

(defun clk (x)
         (clk+ 2
               (loop-clk x)))

(defun test-program (x y)
         (let* ((si (make-state 0 (list x y 0) nil *pi*))
                (k (clk x))
                (sf-prev (m1 si (- k 1)))
                (sf (step sf-prev)))
           `((:clk ,k)
             (:just-enough? ,(and (not (haltedp sf-prev))
                                  (haltedp sf)))
             (:tos ,(top (stack sf))))))

(test-program 5 7)
(test-program 7 5)
(test-program 700 500)

(quote (back to talk))

(thm
      (implies (and (natp x) (natp y) (natp a)
                    (not (zp x)))
               (equal (m1 (make-state 2
                                      (list x y a)
                                      nil
                                      *pi*)
                          11)
                      xxx)))

(quote (back to talk))

(thm
      (implies (and (natp i)
                    (natp j))
               (equal (m1 s (+ i j))
                      (m1 (m1 s i) j)))
      :hints (("Goal" :in-theory (ENABLE M1 STEP))))

(thm
 (implies (and (natp i)
               (natp j))
          (equal (m1 s (+ i j))
                 (m1 (m1 s i) j)))
 :hints (("Goal" :in-theory (ENABLE M1))))

(quote (back to talk))

(defun g (x y a)
       (if (zp x)
           a
           (g (- x 1) y (+ y a))))

(defthm loop-code-correct
       (implies (and (natp x) (natp y) (natp a))
                (equal (m1 (make-state 2
                                       (list x y a)
                                       nil
                                       *pi*)
                           (loop-clk x))
                       (make-state 14
                                   (list 0 y (g x y a))
                                   (push (g x y a) nil)
                                   *pi*))))

(in-theory (disable loop-clk))

(defthm entry-code-correct
       (implies (and (natp x) (natp y))
                (equal (m1 (make-state 0
                                       (list x y)
                                       nil
                                       *pi*)
                           (clk x))
                       (make-state 14
                                   (list 0 y (g x y 0))
                                   (push (g x y 0) nil)
                                   *pi*))))

(in-theory (disable clk))

(quote (back to talk))

(defthm lemma
       (implies (and (natp x) (natp y) (natp a))
                (equal (g x y a)
                       (+ a (* x y)))))

(defthm algorithm-implements-spec
       (implies (and (natp x) (natp y))
                (equal (g x y 0)
                       (* x y))))

(defthm main-goal
       (implies (and (natp x)
                     (natp y)
                     (equal si (make-state 0 (list x y) nil *pi*))
                     (equal sf (m1 si (clk x))))
                (and (haltedp sf)
                     (equal (top (stack sf)) (* x y))))
       :rule-classes nil)

(quote (the end))
